import numpy as np
from qibolab import AcquisitionType, AveragingMode, ExecutionParameters
from qibolab.platform import Platform
from qibolab.pulses import PulseSequence, PulseType
from qibolab.qubits import QubitPairId
from qibolab.sweeper import Parameter, Sweeper, SweeperType

from qibocal.auto.operation import Results, Routine

from ..two_qubit_interaction.chevron.chevron import (
    ChevronData,
    ChevronParameters,
    _plot,
)
from ..two_qubit_interaction.utils import order_pair


def _acquisition(
    params: ChevronParameters,
    platform: Platform,
    targets: list[QubitPairId],
) -> ChevronData:
    r"""
    Perform an CZ experiment between pairs of qubits by changing the coupler state,
    qubits need to be pulses into their interaction point.

    Args:
        platform: Platform to use.
        params: Experiment parameters.
        targets (list): List of pairs to use sequentially.

    Returns:
        ChevronData: Acquisition data.
    """
    # define the parameter to sweep and its range:
    delta_amplitude_range = np.arange(
        params.amplitude_min_factor,
        params.amplitude_max_factor,
        params.amplitude_step_factor,
    )
    delta_duration_range = np.arange(
        params.duration_min, params.duration_max, params.duration_step
    )

    data = ChevronData()
    for pair in targets:
        sequence = PulseSequence()

        ordered_pair = order_pair(pair, platform)

        # initialize system to state 11(CZ) or 10(iSWAP)
        if params.native == "CZ":
            initialize_lowfreq = platform.create_RX_pulse(ordered_pair[0], start=0)
            sequence.add(initialize_lowfreq)

        initialize_highfreq = platform.create_RX_pulse(ordered_pair[1], start=0)

        sequence.add(initialize_highfreq)

        if params.native == "CZ":
            native_gate, _ = platform.create_CZ_pulse_sequence(
                (ordered_pair[1], ordered_pair[0]),
                start=sequence.finish + params.dt,
            )
        elif params.native == "iSWAP":
            native_gate, _ = platform.create_iSWAP_pulse_sequence(
                (ordered_pair[1], ordered_pair[0]),
                start=sequence.finish + params.dt,
            )
        data.native_amplitude[ordered_pair] = getattr(
            native_gate.coupler_pulses(*pair)[:1][0], "amplitude"
        )
        sequence.add(native_gate)

        ro_pulse1 = platform.create_MZ_pulse(
            ordered_pair[1], start=sequence.finish + params.dt
        )
        ro_pulse2 = platform.create_MZ_pulse(
            ordered_pair[0], start=sequence.finish + params.dt
        )

        sequence += ro_pulse1 + ro_pulse2

        coupler_flux_pulses = [p for p in native_gate.coupler_pulses(*pair)]
        assert (
            len(coupler_flux_pulses) == 1
        ), f"coupler_chevron expects exactly one coupler flux pulse, but {len(coupler_flux_pulses)} are present."
        qubit_flux_pulses = [
            p for p in native_gate.get_qubit_pulses(*pair) if p.type is PulseType.FLUX
        ]
        assert all(
            len(list(filter(lambda x: x.qubit == q, qubit_flux_pulses))) < 2
            for q in pair
        ), f"coupler_chevron expects no more than 1 flux pulse for each qubit, but more are present for the pair {pair}"
        sweeper_amplitude = Sweeper(
            Parameter.amplitude,
            delta_amplitude_range,
            pulses=coupler_flux_pulses,
            type=SweeperType.FACTOR,
        )
        sweeper_duration = Sweeper(
            Parameter.duration,
            delta_duration_range,
            pulses=coupler_flux_pulses + qubit_flux_pulses,
        )

        results = platform.sweep(
            sequence,
            ExecutionParameters(
                nshots=params.nshots,
                acquisition_type=AcquisitionType.DISCRIMINATION,
                averaging_mode=AveragingMode.CYCLIC,
            ),
            sweeper_duration,
            sweeper_amplitude,
        )

        data.register_qubit(
            ordered_pair[0],
            ordered_pair[1],
            delta_duration_range,
            delta_amplitude_range * data.native_amplitude[ordered_pair],
            results[ordered_pair[0]].probability(state=1),
            results[ordered_pair[1]].probability(state=1),
        )
    data.label = "Probability of state |1>"

    return data


def _fit(data: ChevronData) -> Results:
    """Results for ChevronCouplers."""
    return Results()


def plot(data: ChevronData, fit: Results, target):
    return _plot(data, None, target)


coupler_chevron = Routine(_acquisition, _fit, plot, two_qubit_gates=True)
"""Coupler cz/swap flux routine."""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import ExecutionParameters
from qibolab.platform import Platform
from qibolab.qubits import QubitId

from qibocal.auto.operation import Data, Parameters, Results, Routine

from ...readout_mitigation_matrix import readout_mitigation_matrix
from ...utils import STRING_TYPE, calculate_frequencies
from .pulses import create_mermin_sequences
from .utils import (
    compute_mermin,
    get_mermin_coefficients,
    get_mermin_polynomial,
    get_readout_basis,
)

PLOT_PADDING = 0.2


@dataclass
class MerminParameters(Parameters):
    """Mermin experiment input parameters."""

    ntheta: int
    """Number of angles probed linearly between 0 and 2 pi."""
    native: Optional[bool] = False
    """If True a circuit will be created using only GPI2 and CZ gates."""
    apply_error_mitigation: Optional[bool] = False
    """Error mitigation model"""


MerminType = np.dtype(
    [
        ("theta", float),
        ("basis", STRING_TYPE),
        ("state", int),
        ("frequency", int),
    ]
)


@dataclass
class MerminData(Data):
    """Mermin Data structure."""

    thetas: list
    """Angles probed."""
    data: dict[list[QubitId], npt.NDArray[MerminType]] = field(default_factory=dict)
    """Raw data acquired."""
    mitigation_matrix: dict[list[QubitId], npt.NDArray[np.float64]] = field(
        default_factory=dict
    )
    """Mitigation matrix computed using the readout_mitigation_matrix protocol."""

    @property
    def targets(self):
        return list(self.data)


@dataclass
class MerminResults(Results):
    """Mermin Results class."""

    mermin: dict[tuple[QubitId, ...], npt.NDArray[np.float64]] = field(
        default_factory=dict
    )
    """Raw Mermin value."""

    mermin_mitigated: dict[tuple[QubitId, ...], npt.NDArray[np.float64]] = field(
        default_factory=dict
    )
    """Mitigated Mermin value."""


def _acquisition(
    params: MerminParameters,
    platform: Platform,
    targets: list[list[QubitId]],
) -> MerminData:
    """Data acquisition for Mermin protocol using pulse sequences."""

    thetas = np.linspace(0, 2 * np.pi, params.ntheta)
    data = MerminData(thetas=thetas.tolist())
    if params.apply_error_mitigation:
        mitigation_data, _ = readout_mitigation_matrix.acquisition(
            readout_mitigation_matrix.parameters_type.load(dict(nshots=params.nshots)),
            platform,
            targets,
        )

        mitigation_results, _ = readout_mitigation_matrix.fit(mitigation_data)
        data.mitigation_matrix = mitigation_results.readout_mitigation_matrix
    platform.connect()
    for qubits in targets:
        mermin_polynomial = get_mermin_polynomial(len(qubits))
        readout_basis = get_readout_basis(mermin_polynomial)

        for theta in thetas:
            mermin_sequences = create_mermin_sequences(
                platform, qubits, readout_basis=readout_basis, theta=theta
            )
            options = ExecutionParameters(nshots=params.nshots)
            # TODO: use unrolling
            for basis, sequence in mermin_sequences.items():
                results = platform.execute_pulse_sequence(sequence, options=options)
                frequencies = calculate_frequencies(results, qubits)
                for state, frequency in enumerate(frequencies.values()):
                    data.register_qubit(
                        MerminType,
                        tuple(qubits),
                        dict(
                            theta=np.array([theta]),
                            basis=np.array([basis]),
                            state=np.array([state]),
                            frequency=np.array([frequency]),
                        ),
                    )
    return data


def _fit(data: MerminData) -> MerminResults:
    """Fitting for Mermin protocol."""
    targets = data.targets
    results = {qubits: [] for qubits in targets}
    mitigated_results = {qubits: [] for qubits in targets}
    basis = np.unique(data.data[targets[0]].basis)
    for qubits in targets:
        mermin_polynomial = get_mermin_polynomial(len(qubits))
        mermin_coefficients = get_mermin_coefficients(mermin_polynomial)

        for theta in data.thetas:
            qubit_data = data.data[qubits]
            outputs = []
            mitigated_outputs = []
            for base in basis:
                frequencies = np.zeros(2 ** len(qubits))
                data_filter = (qubit_data.basis == base) & (qubit_data.theta == theta)
                filtered_data = qubit_data[data_filter]
                state_freq = qubit_data[data_filter].frequency
                for state, freq in zip(filtered_data.state, filtered_data.frequency):
                    frequencies[state] = freq

                outputs.append(
                    {
                        format(i, f"0{len(qubits)}b"): freq
                        for i, freq in enumerate(state_freq)
                    }
                )

                if data.mitigation_matrix:
                    mitigated_output = np.dot(
                        data.mitigation_matrix[qubits],
                        frequencies,
                    )
                    mitigated_outputs.append(
                        {
                            format(i, f"0{len(qubits)}b"): freq
                            for i, freq in enumerate(mitigated_output)
                        }
                    )
            if data.mitigation_matrix:
                mitigated_results[tuple(qubits)].append(
                    compute_mermin(mitigated_outputs, mermin_coefficients)
                )
            results[tuple(qubits)].append(compute_mermin(outputs, mermin_coefficients))
    return MerminResults(
        mermin=results,
        mermin_mitigated=mitigated_results,
    )


def _plot(data: MerminData, fit: MerminResults, target):
    """Plotting function for Mermin protocol."""
    figures = []

    n_qubits = len(target)
    classical_bound = 2 ** (n_qubits // 2)
    quantum_bound = 2 ** ((n_qubits - 1) / 2) * (2 ** (n_qubits // 2))

    fig = go.Figure(
        layout_yaxis_range=[-quantum_bound - PLOT_PADDING, quantum_bound + PLOT_PADDING]
    )
    if fit is not None:
        fig.add_trace(
            go.Scatter(
                x=data.thetas,
                y=fit.mermin[tuple(target)],
                name="Bare",
            )
        )
        if fit.mermin_mitigated:
            fig.add_trace(
                go.Scatter(
                    x=data.thetas,
                    y=fit.mermin_mitigated[tuple(target)],
                    name="Mitigated",
                )
            )

    fig.add_trace(
        go.Scatter(
            mode="lines",
            x=data.thetas,
            y=[+classical_bound] * len(data.thetas),
            line_color="gray",
            name="Classical limit",
            line_dash="dash",
            legendgroup="classic",
        )
    )

    fig.add_trace(
        go.Scatter(
            mode="lines",
            x=data.thetas,
            y=[-classical_bound] * len(data.thetas),
            line_color="gray",
            name="Classical limit",
            legendgroup="classic",
            line_dash="dash",
            showlegend=False,
        )
    )

    fig.add_trace(
        go.Scatter(
            mode="lines",
            x=data.thetas,
            y=[+quantum_bound] * len(data.thetas),
            line_color="gray",
            name="Quantum limit",
            legendgroup="quantum",
        )
    )

    fig.add_trace(
        go.Scatter(
            mode="lines",
            x=data.thetas,
            y=[-quantum_bound] * len(data.thetas),
            line_color="gray",
            name="Quantum limit",
            legendgroup="quantum",
            showlegend=False,
        )
    )

    fig.update_layout(
        xaxis_title="Theta [rad]",
        yaxis_title="Mermin polynomial value",
        xaxis=dict(range=[min(data.thetas), max(data.thetas)]),
    )
    figures.append(fig)

    return figures, ""


mermin = Routine(_acquisition, _fit, _plot)



from qibo.hamiltonians import SymbolicHamiltonian
from qibo.symbols import X, Y


def compute_mermin(frequencies, mermin_coefficients):
    """Computes the chsh inequality out of the frequencies of the 4 circuits executed."""
    assert len(frequencies) == len(mermin_coefficients)
    m = 0
    for j, freq in enumerate(frequencies):
        for key in freq:
            m += (
                mermin_coefficients[j]
                * freq[key]
                * (-1) ** (sum([int(key[k]) for k in range(len(key))]))
            )
    nshots = sum(freq[x] for x in freq)
    if nshots != 0:
        return float(m / nshots)

    return 0


def get_mermin_polynomial(n):
    assert n > 1
    m0 = X(0)
    m0p = Y(0)
    for i in range(1, n):
        mn = m0 * (X(i) + Y(i)) + m0p * (X(i) - Y(i))
        mnp = m0 * (Y(i) - X(i)) + m0p * (X(i) + Y(i))
        m0 = mn.expand()
        m0p = mnp.expand()
    m = m0 / 2 ** ((n - 1) // 2)
    return SymbolicHamiltonian(m.expand())


def get_readout_basis(mermin_polynomial):
    return [
        "".join([factor.name[0] for factor in term.factors])
        for term in mermin_polynomial.terms
    ]


def get_mermin_coefficients(mermin_polynomial):
    return [term.coefficient.real for term in mermin_polynomial.terms]



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
from qibolab import AcquisitionType, AveragingMode, ExecutionParameters
from qibolab.platform import Platform
from qibolab.pulses import PulseSequence
from qibolab.qubits import QubitId
from qibolab.sweeper import Parameter, Sweeper, SweeperType
from scipy.optimize import curve_fit

from ... import update
from ...auto.operation import Data, Parameters, Results, Routine
from ...config import log
from ..utils import GHZ_TO_HZ, HZ_TO_GHZ, extract_feature, table_dict, table_html
from . import utils


@dataclass
class ResonatorFluxParameters(Parameters):
    """ResonatorFlux runcard inputs."""

    freq_width: int
    """Width for frequency sweep relative to the readout frequency [Hz]."""
    freq_step: int
    """Frequency step for sweep [Hz]."""
    bias_width: Optional[float] = None
    """Width for bias sweep [V]."""
    bias_step: Optional[float] = None
    """Bias step for sweep [a.u.]."""


@dataclass
class ResonatorFluxResults(Results):
    """ResonatoFlux outputs."""

    resonator_freq: dict[QubitId, float] = field(default_factory=dict)
    """Readout frequency."""
    coupling: dict[QubitId, float] = field(default_factory=dict)
    """Qubit-resonator coupling."""
    asymmetry: dict[QubitId, float] = field(default_factory=dict)
    """Asymmetry between junctions."""
    sweetspot: dict[QubitId, float] = field(default_factory=dict)
    """Sweetspot for each qubit."""
    matrix_element: dict[QubitId, float] = field(default_factory=dict)
    """Sweetspot for each qubit."""
    fitted_parameters: dict[QubitId, float] = field(default_factory=dict)


ResFluxType = np.dtype(
    [
        ("freq", np.float64),
        ("bias", np.float64),
        ("signal", np.float64),
        ("phase", np.float64),
    ]
)
"""Custom dtype for resonator flux dependence."""


@dataclass
class ResonatorFluxData(Data):
    """ResonatorFlux acquisition outputs."""

    resonator_type: str
    """Resonator type."""
    qubit_frequency: dict[QubitId, float] = field(default_factory=dict)
    """Qubit frequencies."""
    bare_resonator_frequency: dict[QubitId, int] = field(default_factory=dict)
    """Qubit bare resonator frequency power provided by the user."""
    charging_energy: dict[QubitId, float] = field(default_factory=dict)
    """Qubit charging energy in Hz."""

    data: dict[QubitId, npt.NDArray[ResFluxType]] = field(default_factory=dict)
    """Raw data acquired."""

    def register_qubit(self, qubit, freq, bias, signal, phase):
        """Store output for single qubit."""
        self.data[qubit] = utils.create_data_array(
            freq, bias, signal, phase, dtype=ResFluxType
        )


def _acquisition(
    params: ResonatorFluxParameters, platform: Platform, targets: list[QubitId]
) -> ResonatorFluxData:
    """Data acquisition for ResonatorFlux experiment."""

    sequence = PulseSequence()
    ro_pulses = {}
    qubit_frequency = {}
    bare_resonator_frequency = {}
    charging_energy = {}
    for qubit in targets:
        qubit_frequency[qubit] = platform.qubits[qubit].drive_frequency
        bare_resonator_frequency[qubit] = platform.qubits[
            qubit
        ].bare_resonator_frequency
        charging_energy[qubit] = -platform.qubits[qubit].anharmonicity
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    delta_frequency_range = np.arange(
        -params.freq_width / 2, params.freq_width / 2, params.freq_step
    )
    freq_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        [ro_pulses[qubit] for qubit in targets],
        type=SweeperType.OFFSET,
    )

    delta_bias_range = np.arange(
        -params.bias_width / 2, params.bias_width / 2, params.bias_step
    )
    sweepers = [
        Sweeper(
            Parameter.bias,
            delta_bias_range,
            qubits=[platform.qubits[qubit] for qubit in targets],
            type=SweeperType.OFFSET,
        )
    ]

    data = ResonatorFluxData(
        resonator_type=platform.resonator_type,
        qubit_frequency=qubit_frequency,
        bare_resonator_frequency=bare_resonator_frequency,
        charging_energy=charging_energy,
    )
    options = ExecutionParameters(
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=AveragingMode.CYCLIC,
    )
    for bias_sweeper in sweepers:
        results = platform.sweep(sequence, options, bias_sweeper, freq_sweeper)
        # retrieve the results for every qubit
        for qubit in targets:
            result = results[ro_pulses[qubit].serial]
            sweetspot = platform.qubits[qubit].sweetspot
            data.register_qubit(
                qubit,
                signal=result.magnitude,
                phase=result.phase,
                freq=delta_frequency_range + ro_pulses[qubit].frequency,
                bias=delta_bias_range + sweetspot,
            )
    return data


def _fit(data: ResonatorFluxData) -> ResonatorFluxResults:
    """PostProcessing for resonator_flux protocol.

    After applying a mask on the 2D data, the signal is fitted using
    the expected resonator_freq vs flux behavior.
    The fitting procedure requires the knowledge of the bare resonator frequency,
    the charging energy Ec and the maximum qubit frequency which is assumed to be
    the frequency at which the qubit is placed.
    The protocol aims at extracting the sweetspot, the flux coefficient, the coupling,
    the asymmetry and the dressed resonator frequency.
    """

    coupling = {}
    resonator_freq = {}
    asymmetry = {}
    fitted_parameters = {}
    sweetspot = {}
    matrix_element = {}

    for qubit in data.qubits:
        qubit_data = data[qubit]
        biases = qubit_data.bias
        frequencies = qubit_data.freq
        signal = qubit_data.signal

        # extract signal from 2D plot based on SNR mask
        frequencies, biases = extract_feature(
            frequencies, biases, signal, "min" if data.resonator_type == "2D" else "max"
        )

        # define fit function
        def fit_function(
            x: float, g: float, d: float, offset: float, normalization: float
        ):
            """Fit function for resonator flux dependence."""
            return utils.transmon_readout_frequency(
                xi=x,
                w_max=data.qubit_frequency[qubit] * HZ_TO_GHZ,
                xj=0,
                d=d,
                normalization=normalization,
                offset=offset,
                crosstalk_element=1,
                charging_energy=data.charging_energy[qubit] * HZ_TO_GHZ,
                resonator_freq=data.bare_resonator_frequency[qubit] * HZ_TO_GHZ,
                g=g,
            )

        try:
            popt, _ = curve_fit(
                fit_function,
                biases,
                frequencies * HZ_TO_GHZ,
                bounds=(
                    [0, 0, -1, 0.5],
                    [0.5, 1, 1, +1],
                ),
                maxfev=100000,
            )
            fitted_parameters[qubit] = {
                "w_max": data.qubit_frequency[qubit] * HZ_TO_GHZ,
                "xj": 0,
                "d": popt[1],
                "normalization": popt[3],
                "offset": popt[2],
                "crosstalk_element": 1,
                "charging_energy": data.charging_energy[qubit] * HZ_TO_GHZ,
                "resonator_freq": data.bare_resonator_frequency[qubit] * HZ_TO_GHZ,
                "g": popt[0],
            }
            matrix_element[qubit] = popt[3]
            sweetspot[qubit] = (np.round(popt[2]) - popt[2]) / popt[3]
            resonator_freq[qubit] = fit_function(sweetspot[qubit], *popt) * GHZ_TO_HZ
            coupling[qubit] = popt[0]
            asymmetry[qubit] = popt[1]
        except ValueError as e:
            log.error(
                f"Error in resonator_flux protocol fit: {e} "
                "The threshold for the SNR mask is probably too high. "
                "Lowering the value of `threshold` in `extract_*_feature`"
                "should fix the problem."
            )
    return ResonatorFluxResults(
        resonator_freq=resonator_freq,
        coupling=coupling,
        matrix_element=matrix_element,
        sweetspot=sweetspot,
        asymmetry=asymmetry,
        fitted_parameters=fitted_parameters,
    )


def _plot(data: ResonatorFluxData, fit: ResonatorFluxResults, target: QubitId):
    """Plotting function for ResonatorFlux Experiment."""
    figures = utils.flux_dependence_plot(
        data, fit, target, utils.transmon_readout_frequency
    )

    if fit is not None:
        fitting_report = table_html(
            table_dict(
                target,
                [
                    "Coupling g [MHz]",
                    "Dressed resonator freq [Hz]",
                    "Asymmetry",
                    "Sweetspot [V]",
                    "Flux dependence [V]^-1",
                    "Chi [MHz]",
                ],
                [
                    np.round(fit.coupling[target] * 1e3, 2),
                    np.round(fit.resonator_freq[target], 6),
                    np.round(fit.asymmetry[target], 3),
                    np.round(fit.sweetspot[target], 4),
                    np.round(fit.matrix_element[target], 4),
                    np.round(
                        (
                            data.bare_resonator_frequency[target]
                            - fit.resonator_freq[target]
                        )
                        * 1e-6,
                        2,
                    ),
                ],
            )
        )
        return figures, fitting_report
    return figures, ""


def _update(results: ResonatorFluxResults, platform: Platform, qubit: QubitId):
    update.readout_frequency(results.resonator_freq[qubit], platform, qubit)
    update.coupling(results.coupling[qubit], platform, qubit)
    update.asymmetry(results.coupling[qubit], platform, qubit)
    update.sweetspot(results.sweetspot[qubit], platform, qubit)
    update.crosstalk_matrix(results.matrix_element[qubit], platform, qubit, qubit)


resonator_flux = Routine(_acquisition, _fit, _plot, _update)
"""ResonatorFlux Routine object."""



from collections import defaultdict

import numpy as np
from qibolab.pulses import PulseSequence


def create_mermin_sequence(platform, qubits, theta=None):
    """Creates the pulse sequence to generate the bell states and with a theta-measurement"""

    nqubits = len(qubits)
    if theta is None:
        theta = ((nqubits - 1) * 0.25 * np.pi) % (2 * np.pi)

    virtual_z_phases = defaultdict(int)
    sequence = PulseSequence()

    for qubit in qubits:
        sequence.add(
            platform.create_RX90_pulse(
                qubit, start=0, relative_phase=virtual_z_phases[qubit] + np.pi / 2
            )
        )

    # TODO: Not hardcode topology

    # qubits[0] needs to be the center qubit where everything is connected
    for i in range(1, len(qubits)):
        (cz_sequence1, cz_virtual_z_phases) = platform.create_CZ_pulse_sequence(
            [qubits[0]] + [qubits[i]], sequence.finish + 8  # TODO: ask for the 8
        )
        sequence.add(cz_sequence1)
        for qubit in cz_virtual_z_phases:
            virtual_z_phases[qubit] += cz_virtual_z_phases[qubit]

    t = sequence.finish + 8

    for i in range(1, len(qubits)):
        sequence.add(
            platform.create_RX90_pulse(
                qubits[i],
                start=t,
                relative_phase=virtual_z_phases[qubits[i]] - np.pi / 2,
            )
        )

    virtual_z_phases[qubits[0]] -= theta

    return sequence, virtual_z_phases


def create_mermin_sequences(platform, qubits, readout_basis, theta):
    """Creates the pulse sequences needed for the 4 measurement settings for chsh."""

    mermin_sequences = {}

    for basis in readout_basis:
        sequence, virtual_z_phases = create_mermin_sequence(
            platform, qubits, theta=theta
        )
        # t = sequence.finish
        for i, base in enumerate(basis):
            if base == "X":
                sequence.add(
                    platform.create_RX90_pulse(
                        qubits[i],
                        start=sequence.finish,
                        relative_phase=virtual_z_phases[qubits[i]] + np.pi / 2,
                    )
                )
            if base == "Y":
                sequence.add(
                    platform.create_RX90_pulse(
                        qubits[i],
                        start=sequence.finish,
                        relative_phase=virtual_z_phases[qubits[i]],
                    )
                )
        measurement_start = sequence.finish

        for qubit in qubits:
            sequence.add(platform.create_MZ_pulse(qubit, start=measurement_start))

        mermin_sequences[basis] = sequence
    return mermin_sequences



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
import plotly.express as px
from qibo import gates
from qibo.backends import get_backend
from qibo.models import Circuit
from qibolab.platform import Platform
from qibolab.qubits import QubitId

from qibocal.auto.operation import Data, Parameters, Results, Routine
from qibocal.auto.transpile import dummy_transpiler, execute_transpiled_circuit
from qibocal.config import log


@dataclass
class ReadoutMitigationMatrixParameters(Parameters):
    """ReadoutMitigationMatrix matrix inputs."""

    nshots: Optional[int] = None
    """Number of shots."""
    relaxation_time: Optional[int] = None
    """Relaxation time [ns]."""


@dataclass
class ReadoutMitigationMatrixResults(Results):
    readout_mitigation_matrix: dict[tuple[QubitId, ...], npt.NDArray[np.float64]] = (
        field(default_factory=dict)
    )
    """Readout mitigation matrices (inverse of measurement matrix)."""


ReadoutMitigationMatrixType = np.dtype(
    [
        ("state", int),
        ("frequency", np.float64),
    ]
)


@dataclass
class ReadoutMitigationMatrixData(Data):
    """ReadoutMitigationMatrix acquisition outputs."""

    qubit_list: list[QubitId]
    """List of qubit ids"""
    nshots: int
    """Number of shots"""
    data: dict = field(default_factory=dict)
    """Raw data acquited."""


def _acquisition(
    params: ReadoutMitigationMatrixParameters,
    platform: Platform,
    targets: list[list[QubitId]],
) -> ReadoutMitigationMatrixData:
    data = ReadoutMitigationMatrixData(
        nshots=params.nshots, qubit_list=[list(qq) for qq in targets]
    )
    backend = get_backend()
    backend.platform = platform
    transpiler = dummy_transpiler(backend)
    qubit_map = [i for i in range(platform.nqubits)]
    for qubits in targets:
        nqubits = len(qubits)
        for i in range(2**nqubits):
            state = format(i, f"0{nqubits}b")
            c = Circuit(
                nqubits,
            )
            for q, bit in enumerate(state):
                if bit == "1":
                    c.add(gates.X(q))
            c.add(gates.M(*range(nqubits)))
            _, results = execute_transpiled_circuit(
                c, qubits, backend, nshots=params.nshots, transpiler=transpiler
            )
            frequencies = np.zeros(2 ** len(qubits))
            for i, freq in results.frequencies().items():
                frequencies[int(i, 2)] = freq
            for freq in frequencies:
                data.register_qubit(
                    ReadoutMitigationMatrixType,
                    (qubits),
                    dict(
                        state=np.array([int(state, 2)]),
                        frequency=freq,
                    ),
                )
    return data


def _fit(data: ReadoutMitigationMatrixData) -> ReadoutMitigationMatrixResults:
    """Post processing for readout mitigation matrix protocol."""
    readout_mitigation_matrix = {}
    for qubits in data.qubit_list:
        qubit_data = data.data[tuple(qubits)]
        mitigation_matrix = []
        for state in range(2 ** len(qubits)):
            mitigation_matrix.append(qubit_data[qubit_data.state == state].frequency)
        mitigation_matrix = np.vstack(mitigation_matrix) / data.nshots
        try:
            readout_mitigation_matrix[tuple(qubits)] = np.linalg.inv(
                mitigation_matrix
            ).tolist()
        except np.linalg.LinAlgError as e:
            log.warning(f"ReadoutMitigationMatrix: the fitting was not succesful. {e}")
    res = ReadoutMitigationMatrixResults(
        readout_mitigation_matrix=readout_mitigation_matrix,
    )

    return res


def _plot(
    data: ReadoutMitigationMatrixData,
    fit: ReadoutMitigationMatrixResults,
    target: list[QubitId],
):
    """Plotting function for readout mitigation matrix."""
    fitting_report = ""
    figs = []
    if fit is not None:
        computational_basis = [
            format(i, f"0{len(target)}b") for i in range(2 ** len(target))
        ]
        measurement_matrix = np.linalg.inv(fit.readout_mitigation_matrix[tuple(target)])
        z = measurement_matrix
        fig = px.imshow(
            z,
            x=computational_basis,
            y=computational_basis,
            text_auto=True,
            labels={
                "x": "Prepeared States",
                "y": "Measured States",
                "color": "Probabilities",
            },
            width=700,
            height=700,
        )
        figs.append(fig)
    return figs, fitting_report


readout_mitigation_matrix = Routine(_acquisition, _fit, _plot)
"""Readout mitigation matrix protocol."""



from dataclasses import dataclass, field, fields
from typing import Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, ExecutionParameters
from qibolab.platform import Platform
from qibolab.pulses import PulseSequence
from qibolab.qubits import QubitId
from qibolab.sweeper import Parameter, Sweeper, SweeperType

from ...auto.operation import Routine
from ...config import log
from ..utils import table_dict, table_html
from .ramsey import (
    COLORBAND,
    COLORBAND_LINE,
    RamseySignalData,
    RamseySignalParameters,
    RamseySignalResults,
    RamseyType,
    _update,
)
from .utils import fitting, process_fit, ramsey_fit, ramsey_sequence


@dataclass
class RamseyZZParameters(RamseySignalParameters):
    """RamseyZZ runcard inputs."""

    target_qubit: Optional[QubitId] = None
    """Target qubit that will be excited."""


@dataclass
class RamseyZZResults(RamseySignalResults):
    """RamseyZZ outputs."""

    def __contains__(self, qubit: QubitId):
        # TODO: to be improved
        return all(
            list(getattr(self, field.name))[0][0] == qubit
            for field in fields(self)
            if isinstance(getattr(self, field.name), dict)
        )


@dataclass
class RamseyZZData(RamseySignalData):
    """RamseyZZ acquisition outputs."""

    target_qubit: Optional[QubitId] = None
    """Qubit that will be excited."""
    data: dict[tuple[QubitId, str], npt.NDArray[RamseyType]] = field(
        default_factory=dict
    )
    """Raw data acquired."""


def _acquisition(
    params: RamseyZZParameters,
    platform: Platform,
    targets: list[QubitId],
) -> RamseyZZData:
    """Data acquisition for RamseyZZ Experiment.

    Standard Ramsey experiment repeated twice.
    In the second execution one qubit is brought to the excited state.
    """

    waits = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    options = ExecutionParameters(
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.SINGLESHOT,
    )

    data = RamseyZZData(
        detuning=params.detuning,
        qubit_freqs={
            qubit: platform.qubits[qubit].native_gates.RX.frequency for qubit in targets
        },
        target_qubit=params.target_qubit,
    )

    for setup in ["I", "X"]:
        if not params.unrolling:
            sequence = PulseSequence()
            for qubit in targets:
                sequence += ramsey_sequence(
                    platform=platform,
                    qubit=qubit,
                    detuning=params.detuning,
                    target_qubit=params.target_qubit if setup == "X" else None,
                )

            sweeper = Sweeper(
                Parameter.start,
                waits,
                [sequence.get_qubit_pulses(qubit).qd_pulses[-1] for qubit in targets],
                type=SweeperType.ABSOLUTE,
            )

            # execute the sweep
            results = platform.sweep(
                sequence,
                options,
                sweeper,
            )

            for qubit in targets:
                probs = results[qubit].probability(state=1)
                errors = [np.sqrt(prob * (1 - prob) / params.nshots) for prob in probs]

        else:
            sequences, all_ro_pulses = [], []
            probs, errors = [], []
            for wait in waits:
                sequence = PulseSequence()
                for qubit in targets:
                    sequence += ramsey_sequence(
                        platform=platform,
                        qubit=qubit,
                        wait=wait,
                        detuning=params.detuning,
                        target_qubit=params.target_qubit if setup == "X" else None,
                    )

                sequences.append(sequence)
                all_ro_pulses.append(sequence.ro_pulses)

            results = platform.execute_pulse_sequences(sequences, options)

            for wait, ro_pulses in zip(waits, all_ro_pulses):
                for qubit in targets:
                    prob = results[ro_pulses[qubit].serial][0].probability(state=1)
                    probs.append(prob)
                    errors.append(np.sqrt(prob * (1 - prob) / params.nshots))

        for qubit in targets:
            data.register_qubit(
                RamseyType,
                (qubit, setup),
                dict(
                    wait=waits,
                    prob=probs,
                    errors=errors,
                ),
            )

    return data


def _fit(data: RamseyZZData) -> RamseyZZResults:
    """Fitting procedure for RamseyZZ protocol.

    Standard Ramsey fitting procedure is applied for both version of
    the experiment.

    """
    waits = data.waits
    popts = {}
    freq_measure = {}
    t2_measure = {}
    delta_phys_measure = {}
    delta_fitting_measure = {}
    for qubit in data.qubits:
        for setup in ["I", "X"]:
            qubit_data = data[qubit, setup]
            qubit_freq = data.qubit_freqs[qubit]
            probs = qubit_data["prob"]
            try:
                popt, perr = fitting(waits, probs, qubit_data.errors)
                (
                    freq_measure[qubit, setup],
                    t2_measure[qubit, setup],
                    delta_phys_measure[qubit, setup],
                    delta_fitting_measure[qubit, setup],
                    popts[qubit, setup],
                ) = process_fit(popt, perr, qubit_freq, data.detuning)
            except Exception as e:
                log.warning(f"Ramsey fitting failed for qubit {qubit} due to {e}.")
    return RamseyZZResults(
        detuning=data.detuning,
        frequency=freq_measure,
        t2=t2_measure,
        delta_phys=delta_phys_measure,
        delta_fitting=delta_fitting_measure,
        fitted_parameters=popts,
    )


def _plot(data: RamseyZZData, target: QubitId, fit: RamseyZZResults = None):
    """Plotting function for Ramsey Experiment."""

    figures = []
    fitting_report = ""

    waits = data[target, "I"].wait
    probs_I = data.data[target, "I"].prob
    probs_X = data.data[target, "X"].prob

    error_bars_I = data.data[target, "I"].errors
    error_bars_X = data.data[target, "X"].errors
    fig = go.Figure(
        [
            go.Scatter(
                x=waits,
                y=probs_I,
                opacity=1,
                name="I",
                showlegend=True,
                legendgroup="I ",
                mode="lines",
            ),
            go.Scatter(
                x=np.concatenate((waits, waits[::-1])),
                y=np.concatenate(
                    (probs_I + error_bars_I, (probs_I - error_bars_I)[::-1])
                ),
                fill="toself",
                fillcolor=COLORBAND,
                line=dict(color=COLORBAND_LINE),
                showlegend=True,
                name="Errors I",
            ),
            go.Scatter(
                x=waits,
                y=probs_X,
                opacity=1,
                name="X",
                showlegend=True,
                legendgroup="X",
                mode="lines",
            ),
            go.Scatter(
                x=np.concatenate((waits, waits[::-1])),
                y=np.concatenate(
                    (probs_X + error_bars_X, (probs_X - error_bars_X)[::-1])
                ),
                fill="toself",
                fillcolor=COLORBAND,
                line=dict(color=COLORBAND_LINE),
                showlegend=True,
                name="Errors X",
            ),
        ]
    )

    if fit is not None:
        fig.add_trace(
            go.Scatter(
                x=waits,
                y=ramsey_fit(waits, *fit.fitted_parameters[target, "I"]),
                name="Fit I",
                line=go.scatter.Line(dash="dot"),
            )
        )

        fig.add_trace(
            go.Scatter(
                x=waits,
                y=ramsey_fit(waits, *fit.fitted_parameters[target, "X"]),
                name="Fit X",
                line=go.scatter.Line(dash="dot"),
            )
        )
        fitting_report = table_html(
            table_dict(
                data.target_qubit,
                [
                    "ZZ  [kHz]",
                ],
                [
                    np.round(
                        (fit.frequency[target, "X"][0] - fit.frequency[target, "I"][0])
                        * 1e-3,
                        0,
                    ),
                ],
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Time [ns]",
        yaxis_title="Excited state probability",
    )

    figures.append(fig)

    return figures, fitting_report


ramsey_zz = Routine(_acquisition, _fit, _plot, _update)
"""Ramsey Routine object."""
